#include "Main.h"

void MainApp::processMovement(const vec3 &newPosition){
	vec3 pos = newPosition;

	coll.pushSphere(pos, 48);

	position = pos;
}

void MainApp::resetCamera(){
	position = vec3(-390, -120, 390);
	wx = 0;
	wy = -1.9f;
	wz = 0;
}

struct PolyData {
	vec3 v0, vDir0, vDir1;
	vec2 t0, tDir0, tDir1;
};

bool MainApp::init(){
	unsigned int i, j;

	model = new OpenGLModel();
	model->loadFromFile("../Models/Corridor/Map.hmdl");

	for (i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->insertAttribute(ATT_TEXCOORD, ATT_FLOAT, 2, 1);
	}

	TexturePacker texPacker;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			float w = length(src[inds[j + 1]].pos - src[inds[j]].pos) / 18;
			float h = length(src[inds[j + 5]].pos - src[inds[j]].pos) / 18;

			if (w < 8) w = 8;
			if (h < 8) h = 8;

			unsigned int tw = (int) w;
			unsigned int th = (int) h;

			tw = (tw + 4) & ~7;
			th = (th + 4) & ~7;

			texPacker.addRectangle(tw, th);
		}
	}

	unsigned int tw = 256;
	unsigned int th = 256;

	if (!texPacker.assignCoords(&tw, &th)){
#ifdef _WIN32
		MessageBox(NULL, "Lightmap too small", "Error", MB_OK);
#endif
		return false;
	}

	tw = getUpperPowerOfTwo(tw);
	th = getUpperPowerOfTwo(th);

	Set <PolyData *> polyData;
	unsigned int index = 0;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			TextureRectangle *rect = texPacker.getRectangle(index);

			float x0 = float(rect->x + 0.5f) / tw;
			float x1 = float(rect->x + rect->width - 0.5f) / tw;
			float y0 = float(rect->y + 0.5f) / th;
			float y1 = float(rect->y + rect->height - 0.5f) / th;

			src[inds[j    ]].texCoord2 = vec2(x0, y0);
			src[inds[j + 1]].texCoord2 = vec2(x1, y0);
			src[inds[j + 2]].texCoord2 = vec2(x1, y1);
			src[inds[j + 5]].texCoord2 = vec2(x0, y1);

			PolyData *pd = new PolyData;
			pd->v0 = src[inds[j]].pos;
			pd->vDir0 = src[inds[j + 1]].pos - pd->v0;
			pd->vDir1 = src[inds[j + 5]].pos - pd->v0;
			pd->t0 = vec2(x0, y0);
			pd->tDir0 = vec2(x1 - x0, 0);
			pd->tDir1 = vec2(0, y1 - y0);
			polyData.add(pd);

			coll.addTriangle(src[inds[j]].pos, src[inds[j + 1]].pos, src[inds[j + 2]].pos, pd);
			coll.addTriangle(src[inds[j + 3]].pos, src[inds[j + 4]].pos, src[inds[j + 5]].pos, pd);

			index++;
		}
	}
	coll.build();

	lightPos[0] = vec3(800, 0, 750);
	lightSize[0] = 50;
	lightPos[1] = vec3(-400, 0, 160);
	lightSize[1] = 50;


	/*****	Lightmap generation code *****/
/*
	unsigned char *lMap = new unsigned char[tw * th];
	memset(lMap, 0, tw * th);

	vec3 samples[SAMPLE_COUNT];
	for (unsigned int x = 0; x < SAMPLE_COUNT; x++){
		do {
			samples[x] = vec3(float(rand()), float(rand()), float(rand())) * (2.0f / RAND_MAX) - 1.0f;
		} while (dot(samples[x], samples[x]) > 1);
	}

	index = 0;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			TextureRectangle *rect = texPacker.getRectangle(index);

			vec3 dirS = (src[inds[j + 1]].pos - src[inds[j]].pos);
			vec3 dirT = (src[inds[j + 5]].pos - src[inds[j]].pos);

			vec3 pos = src[inds[j]].pos;
			vec3 normal = src[inds[j]].normal;

			for (int t = 0; t < int(rect->height); t++){
				for (int s = 0; s < int(rect->width); s++){
					vec3 samplePos = pos + saturate((float(s) / (rect->width - 1))) * dirS + saturate((float(t) / (rect->height - 1))) * dirT;

					float lightSum = 0.2f;
					for (unsigned int light = 0; light < LIGHT_COUNT; light++){
						vec3 lightVec = lightPos[light] - samplePos;
						float diffuse = dot(normalize(lightVec), normal);

						float sampleSum = 0.0f;
						if (diffuse > 0){
							coll.pushSphere(samplePos, 5);

							for (unsigned int k = 0; k < SAMPLE_COUNT; k++){
								if (!coll.intersect(samplePos, lightPos[light] + samples[k] * lightSize[light])) sampleSum += 1.0f / SAMPLE_COUNT;
							}

							sampleSum *= saturate(diffuse) / (1.0f + 0.000001f * dot(lightVec, lightVec));
						}
						lightSum += sampleSum;
					}
					lMap[((rect->y + t) * tw + (rect->x + s))] = (unsigned char) (255.0f * min(lightSum, 1.0f));
				}
			}

			index++;
		}
	}

	Image image;
	image.loadFromMemory(lMap, tw, th, FORMAT_I8, true, false);
	image.saveImage("LightMap.dds");
*/
	radius = 200;
	blur = 0.15f;

	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}

	for (unsigned int i = 0; i < VIEW_COUNT; i++){
		if ((rt[i] = renderer->addRenderTexture(512, 512, FORMAT_RGBA8, false)) == TEXTURE_NONE) return false;
		if ((lighting[i] = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;
	}

	if ((combine = renderer->addShader("combine.shd")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((light = renderer->addTexture("../Textures/Particle.tga", TEX_CLAMP)) == TEXTURE_NONE) return false;

	if ((base[0] = renderer->addTexture("../Textures/floor_wood_4.dds")) == TEXTURE_NONE) return false;
	if ((base[1] = renderer->addTexture("../Textures/floor_wood_3.dds")) == TEXTURE_NONE) return false;
	if ((base[2] = renderer->addTexture("../Textures/floor6.dds")) == TEXTURE_NONE) return false;

	if ((lightMap = renderer->addTexture("LightMap.dds", TEX_NOMIPMAPPING)) == TEXTURE_NONE) return false;

	model->uploadToVertexBuffer();

	return true;
}

bool MainApp::unload(){
	return true;
}

void MainApp::drawLights(){
	glColor3f(1, 1, 1);

	renderer->setBlending(GL_ONE, GL_ONE);
	renderer->setTextures(light);
	renderer->setMask(COLOR);
	renderer->apply();

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);

	for (unsigned int i = 0; i < LIGHT_COUNT; i++){
		glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3fv(lightPos[i] + lightSize[i] * (-dx + dy));

			glTexCoord2f(1,0);
			glVertex3fv(lightPos[i] + lightSize[i] * (dx + dy));

			glTexCoord2f(1,1);
			glVertex3fv(lightPos[i] + lightSize[i] * (dx - dy));

			glTexCoord2f(0,1);
			glVertex3fv(lightPos[i] + lightSize[i] * (-dx - dy));
		glEnd();
	}
}

void MainApp::drawScene(const int view){
	renderer->changeRenderTarget(rt[view], TEXTURE_NONE);

	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	projection = projectionMatrixX(1.5f, float(height) / float(width), 0.1f, 6000);
	glLoadMatrixf(transpose(projection));

	glClear(/*GL_COLOR_BUFFER_BIT | */GL_DEPTH_BUFFER_BIT);

	float angleX = blur * cosf(float(view) / VIEW_COUNT * 2 * PI) / sqrtf(radius);
	float angleY = blur * sinf(float(view) / VIEW_COUNT * 2 * PI) / sqrtf(radius);

	modelView = translate(vec3(0, 0, radius)) * rotateXY(-wx + angleX, -wy + angleY) * translate(-focusPoint);

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(transpose(modelView));

	// Main pass
	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		renderer->setShader(lighting[view]);
		renderer->setTexture("Base", base[i]);
		renderer->setTexture("LightMap", lightMap);
		renderer->apply();

		((OpenGLBatch *) model->getBatch(i))->draw();
	}

	drawLights();
}

bool MainApp::drawFrame(){
	if (keys[KEY_ADD]){
		blur += 0.3f * frameTime;
		if (blur > 0.3f) blur = 0.3f;
	}
	if (keys[KEY_SUBTRACT]){
		blur -= 0.3f * frameTime;
		if (blur < 0) blur = 0;
	}


	float cosX = cosf(wx), sinX = sinf(wx), cosY = cosf(wy), sinY = sinf(wy);
	vec3 dz(-cosX * sinY, -sinX, cosX * cosY);

	vec3 point;
	if (coll.intersect(position, position + 32768 * dz, &point)){
		float newRadius = length(position - point);
		radius = lerp(newRadius, radius, powf(0.01f, frameTime));
	}

	focusPoint = position + radius * dz;


	for (int j = 0; j < VIEW_COUNT; j++){
		drawScene(j);
	}
	renderer->changeToDefaultRenderTarget();

	renderer->setShader(combine);
	renderer->setTexture("Tex0", rt[0]);
	renderer->setTexture("Tex1", rt[1]);
	renderer->setTexture("Tex2", rt[2]);
	renderer->setTexture("Tex3", rt[3]);
	renderer->setTexture("Tex4", rt[4]);
	renderer->setTexture("Tex5", rt[5]);
	renderer->setTexture("Tex6", rt[6]);
	renderer->setTexture("Tex7", rt[7]);
	renderer->setDepthFunc(DEPTH_NONE);
	renderer->apply();

	glBegin(GL_QUADS);
		glVertex2f(-1,  1);
		glVertex2f( 1,  1);
		glVertex2f( 1, -1);
		glVertex2f(-1, -1);
	glEnd();

	return true;
}

App *app = new MainApp();
